(* Copyright (c) 2015, Thibaut Girka
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY AUTHORS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE. *)

open CorrelatingProgram;;
open CorrelatingProgram.CPStr.LangDefs;;

type ty =
  | IntType
  | ArrayType of ty * int;;

type decl = string * ty;;

let param_to_c ((name, t) : decl) =
  match t with
  | IntType -> "int " ^ name
  | ArrayType (IntType, 0) -> "const int *" ^ name
  | ArrayType _ -> failwith "Arrays not allowed as parameters!"


let rec decl_to_c ((name, t) : decl) =
  match t with
  | IntType -> "int " ^ name
  | ArrayType (t', n) -> (decl_to_c ((name, t'))) ^ "[" ^ (string_of_int n) ^ "]"

let guard_to_c g =
  let rec aux g acc =
    match g with
    | Coq___U03b5_ -> ["GWUT"]
    | S0 Coq___U03b5_ -> "G"::acc
    | S1 Coq___U03b5_ -> "T_G"::acc
    | S0 g' -> aux g' ("0"::acc)
    | S1 g' -> aux g' ("1"::acc)
  in
  String.concat "" (aux g []);;

let guard_conj_to_label gc =
  String.concat "_"
  (List.map (fun (b, g) ->
    (if b then guard_to_c g else "!" ^ (guard_to_c g))) gc);;

let guard_disj_to_label cl =
  String.concat "__v__"
  (List.map guard_conj_to_label cl);;

let guard_conj_to_c gl =
  String.concat "" (List.map (fun (b, g) ->
    "if (" ^ (if b then guard_to_c g else "!" ^ (guard_to_c g)) ^ ") ") gl);;

let rec lval_to_c = function
  | Var x -> x
  | ArrayAccess (x, e) -> (lval_to_c x) ^ "[" ^ (exp_to_c e) ^ "]"
and exp_to_c = function
  | Int n -> string_of_int n
  | Lval x -> lval_to_c x
  | Sum (e, e') -> "(" ^ (exp_to_c e) ^ " + " ^ (exp_to_c e') ^ ")"
  | Diff (e, e') -> "(" ^ (exp_to_c e) ^ " - " ^ (exp_to_c e') ^ ")"
  | Mult (e, e') -> "(" ^ (exp_to_c e) ^ " * " ^ (exp_to_c e') ^ ")"
  | Div (e, e') -> "(" ^ (exp_to_c e) ^ " / " ^ (exp_to_c e') ^ ")";;

let rec bexp_to_c b =
  match b with
  | BTrue -> "1"
  | BFalse -> "0"
  | BAnd (b, b') -> "(" ^ (bexp_to_c b) ^ ") && (" ^ (bexp_to_c b') ^ ")"
  | BOr (b, b') -> "(" ^ (bexp_to_c b) ^ ") || (" ^ (bexp_to_c b') ^ ")"
  | BNot b -> "!(" ^ (bexp_to_c b) ^ ")"
  | BLE (e, e') -> (exp_to_c e) ^ " <= " ^ (exp_to_c e')
  | BLT (e, e') -> (exp_to_c e) ^ " < " ^ (exp_to_c e')
  | BGE (e, e') -> (exp_to_c e) ^ " >= " ^ (exp_to_c e')
  | BGT (e, e') -> (exp_to_c e) ^ " > " ^ (exp_to_c e')
  | BEQ (e, e') -> (exp_to_c e) ^ " == " ^ (exp_to_c e')
  | BNEQ (e, e') -> (exp_to_c e) ^ " != " ^ (exp_to_c e');;

let leaf_to_c (c : cmd_a) =
  match c with
  | Skip -> ""
  | Break -> "break"
  | Continue -> "continue"
  | Assign (x, e) -> (lval_to_c x) ^ " = " ^ (exp_to_c e)

let rec cmd_to_c (indent : string) (c : cmd) =
  match c with
  | Leaf Skip -> ""
  | Leaf a -> indent ^ (leaf_to_c a) ^ ";\n"
  | Seq (c, c') -> (cmd_to_c indent c) ^ (cmd_to_c indent c')
  | IfThenElse (b, c, c') -> indent ^ "if (" ^ (bexp_to_c b) ^ ") {\n" ^
    (cmd_to_c ("  " ^ indent) c) ^ indent ^ "} else {\n" ^
    (cmd_to_c ("  " ^ indent) c') ^ indent ^ "};\n"
  | While (b, c) -> indent ^ "while (" ^ (bexp_to_c b) ^ ") {\n" ^
    (cmd_to_c ("  " ^ indent) c) ^ indent ^ "};\n"

let acmdg_to_c (c : acmd_g) =
  match c with
  | GAGAssign (g, b) ->
    (guard_to_c g) ^ " = " ^ (bexp_to_c b)
  | GAAssign (x, e) ->
    (lval_to_c x) ^ " = " ^ (exp_to_c e);;

let cmdg_to_c (indent : string) (c : cmd_g) =
  let cp_counter = ref (-1) in
  let rec aux indent = function
  | GSkip -> ""
  | GSeq (c, c') ->
    let s = (aux indent c) in
    let s' = (aux indent c') in
    s ^ "\n" ^ s'
  | GAtomic (gl, a) -> indent ^ (guard_conj_to_c gl) ^ (acmdg_to_c a) ^ ";"
  | GWhile (cl, c) ->
    let label = guard_disj_to_label cl in
    let indent = indent ^ "  " in
    label ^ ":\n" ^
    (aux indent c) ^ "\n" ^
    (String.concat "\n"
    (List.map (fun gl -> indent ^ (guard_conj_to_c gl) ^ "goto " ^ label ^ ";") cl))
  | GCPoint ->
    cp_counter := !cp_counter + 1 ;
    "\n" ^ indent ^ "/* Correlating point */\n" ^
    indent ^ "{ char *CP" ^ (string_of_int !cp_counter) ^ "; }"
  in aux indent (GSeq (c, GCPoint));;
