(* Copyright (c) 2015, Thibaut Girka
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY AUTHORS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE. *)

open CorrelatingProgram;;
open CorrelatingProgram.CPStr;;
open CorrelatingProgram.CPStr.LangDefs;;

open Common;;

(* Various useful modules *)

module GuardSet = Set.Make(struct type t = guard let compare = compare end);;
module StringSet = Set.Make(struct type t = string let compare = String.compare end);;

let list_guards c =
  let add_gl acc gl = List.fold_left (fun acc (_, g) ->
    GuardSet.add g acc) acc gl
  in
  let add_cl acc cl = List.fold_left add_gl acc cl in
  let rec aux acc = function
  | GAtomic (gl, a) ->
    let acc = add_gl acc gl in
    begin match a with
    | GAAssign _ -> acc
    | GAGAssign (g, _) -> GuardSet.add g acc
    end
  | GSkip | GCPoint -> acc
  | GSeq (c, c') -> aux (aux acc c) c'
  | GWhile (cl, c) ->
    add_cl (aux acc c) cl
  in
  aux (GuardSet.empty) c

let rec get_var_from_lval = function
  | Var x -> x
  | ArrayAccess (lval, _) -> get_var_from_lval lval

let to_c (cpfun : cmd_diff -> cmd_g)
         (decls, decls')
         (cd : cmd_diff) (name : string) (params : decl list) : string * string =
  let tagged_params = List.map (fun (x, ty) -> (str_t x, ty)) params in
  let tagged_params' = List.map (fun (x, ty) -> (str_t' x, ty)) params in
  let c = cpfun cd in
  let guards = list_guards c in
("/* Stuff needed by cdizy: */
typedef short Guard;

/* External functions, don't include standard libs because cdizy is weird */
extern int atoi (const char *nptr);
extern int printf(const char *format, ...);

/* In order to be able to inspect the result... */
int res, T_res;

void " ^ name ^ "(" ^ (String.concat ", " (List.map param_to_c tagged_params)) ^ ")
{
  /* Declare guards */
" ^ (String.concat "\n" (GuardSet.fold (fun g acc -> ("  Guard " ^ (guard_to_c g) ^ " = 1;")::acc) guards [])) ^ "

  /* Declare/initialize tagged params */
" ^ (String.concat "\n" (List.map2 (fun (a, _) b -> "  " ^ (param_to_c b) ^ " = " ^ a ^ ";") tagged_params tagged_params')) ^ "

  /* Declare other variables */
" ^ (String.concat "\n" (List.map (fun (x, t) -> "  " ^ (decl_to_c (str_t x, t)) ^ ";") decls)) ^ "
" ^ (String.concat "\n" (List.map (fun (x, t) -> "  " ^ (decl_to_c (str_t' x, t)) ^ ";") decls')) ^ "

  /* Here is the body! */
" ^ (cmdg_to_c "  " c) ^ "

  /* Set the output... */
  res = O_result;
  T_res = T_O_result;
}",
"#include <stdio.h>

extern int res, T_res;
extern void " ^ name ^ "(" ^ (String.concat ", " (List.map param_to_c tagged_params)) ^ ");

int main(int argc, const char **argv)
{
  if (argc <= " ^ (string_of_int (List.length params)) ^ ")
  {
    printf(\"Usage: %s " ^ (String.concat " " (List.map fst params)) ^ "\\n\", argv[0]);
    return 1;
  }

  " ^ name ^ "(" ^ (String.concat ", " (List.mapi (fun i _ -> "atoi(argv[" ^ (string_of_int (i + 1)) ^ "])") params)) ^ ");

  printf(\"%d\\n\", res);
  printf(\"%d\\n\", T_res);

  return 0;
}")

let proj_to_c (projfun : cmd_diff -> cmd)
         decls
         (cd : cmd_diff) (name : string) (params : decl list) : string * string =
  let c = projfun cd in
("/* In order to be able to inspect the result... */
int res;

void " ^ name ^ "(" ^ (String.concat ", " (List.map param_to_c params)) ^ ")
{
  /* Declare other variables */
" ^ (String.concat "\n" (List.map (fun x -> "  " ^ (decl_to_c x) ^ ";") decls)) ^ "

  /* Here is the body! */
" ^ (cmd_to_c "  " c) ^ "

  /* Set the output... */
  res = result;
}
",
"#include <stdio.h>

extern int res;
extern void " ^ name ^ "(" ^ (String.concat ", " (List.map param_to_c params)) ^ ");

int main(int argc, const char **argv)
{
  if (argc <= " ^ (string_of_int (List.length params)) ^ ")
  {
    printf(\"Usage: %s " ^ (String.concat " " (List.map fst params)) ^ "\\n\", argv[0]);
    return 1;
  }

  " ^ name ^ "(" ^ (String.concat ", " (List.mapi (fun i _ -> "atoi(argv[" ^ (string_of_int (i + 1)) ^ "])") params)) ^ ");

  printf(\"%d\\n\", res);

  return 0;
}")

let proj_to_imp (projfun : cmd_diff -> cmd)
         (decls : decl list)
         (cd : cmd_diff) (name : string) (params : decl list) : string =
  let c = projfun cd in
"void " ^ name ^ "(" ^ (String.concat ", " (List.map param_to_c params)) ^ ")
{
" ^ (String.concat "\n" (List.map (fun x -> "  " ^ (decl_to_c x) ^ ";") decls)) ^ "

" ^ (cmd_to_c "  " c) ^ "
}
"



let compute_cp output_dummy output_projs output_main insert_cps
               (filename : string) : unit =
  let chopped_ext = Filename.chop_extension filename in
  let dir = Filename.dirname chopped_ext in
  let basename = Filename.basename chopped_ext in
  let output_file prefix str suffix =
    let filename = Filename.concat dir
        (prefix ^ basename ^ "." ^ suffix)
    in
    let out_channel = open_out filename in
    let () = output_string out_channel str in
    close_out out_channel
  in
  let in_channel = open_in filename in
  let lex = Lexing.from_channel in_channel in
  try
    let (name, params, decls_diff, cd) = Parser.function_diffmap Lexer.main lex in
    let cd = if insert_cps then CorrelatingProgram.CPStr.add_cpoints cd else cd in
    let () = close_in in_channel in
    let (result, result_main) = to_c correlating_program0 decls_diff cd name params in
    output_file "union." result "c" ;
    if output_main then output_file "main.union." result_main "c" ;
    let () = if output_dummy then
      let (result, _) = to_c dummy_correlating_program0 decls_diff cd name params in
      output_file "union.seq." result "c"
    in
    if output_projs then
      let (decls, decls') = decls_diff in
      let (result, result_main) = proj_to_c (__U03a0___U2080_) decls cd name params in
      output_file "" result "c" ;
      if output_main then output_file "main." result_main "c" ;
      let (result, result_main) = proj_to_c (__U03a0___U2081_) decls' cd name params in
      output_file "patched." result "c" ;
      if output_main then output_file "main.patched." result_main "c" ;
      let result = proj_to_imp (__U03a0___U2080_) decls cd name params in
      output_file "" result "imp" ;
      let result = proj_to_imp (__U03a0___U2081_) decls' cd name params in
      output_file "patched." result "imp"

  with Parsing.Parse_error ->
    let cpos = lex.Lexing.lex_curr_p in
    Printf.eprintf "Line %d, character %d: Syntax error.\n"
        cpos.Lexing.pos_lnum (cpos.Lexing.pos_cnum - cpos.Lexing.pos_bol);
    Pervasives.exit 1


let main (argv : string array) =
  let output_dummy = ref false in
  let output_projs = ref false in
  let output_main = ref false in
  let insert_cps = ref false in
  let used = ref false in
  let args_spec = [
    ("--enable-seq-cp", (Arg.Set output_dummy),
        "Output a correlating program with the two programs put end to end");
    ("--insert-cps", (Arg.Set insert_cps),
        "Automatically insert Correlating Points in the correlating program");
    ("--enable-projs", (Arg.Set output_projs),
        "Output original and modified program corresponding to the diff");
    ("--enable-main", (Arg.Set output_main),
        "Output a main function in a separate file");
  ] in
  let usage_string = "Usage: correlating_program <options> <files>" in
  let f x =
    used := true ;
    compute_cp (!output_dummy) (!output_projs) (!output_main) (!insert_cps) x in
  try
    Arg.parse_argv argv args_spec f usage_string ;
    if not (!used) then Arg.usage args_spec usage_string
  with
  | Arg.Help s ->
    print_string s
  | Arg.Bad s ->
    prerr_string s

let () = main (Sys.argv);;
