(* Copyright (c) 2015, Thibaut Girka
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY AUTHORS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE. *)

{

  open Lexing
  open Parser

  let is_first_line_char lexbuf =
    let pos = lexeme_start_p lexbuf in
    pos.pos_cnum = pos.pos_bol;;
}

let newline = ('\n' | '\r' | "\r\n")
let whitespace = [ ' ' '\t' ]

rule main = parse
(* Things that really shouldn't be keywords *)
| "const" { CONSTMOD }
| "void" { VOIDTYPE }
| "int" { INTTYPE }
(* Keywords *)
| "if" { IF }
| "else" { ELSE }
| "while" { WHILE }
| "break" { BREAK }
| "continue" { CONTINUE }
| "//" [^ '\r' '\n']* newline
  { new_line lexbuf ; main lexbuf } (* skip comments *)
| whitespace+
  { main lexbuf }
| newline
  { new_line lexbuf ; main lexbuf }
| ";"
  { SEMICOLON }
| "{"
  { LBRACE }
| "}"
  { RBRACE }
| "⟨"
  { LABRACKET }
| "⟩"
  { RABRACKET }
| "["
  { LBRACKET }
| "]"
  { RBRACKET }
| "("
  { LPAREN }
| ")"
  { RPAREN }
| "true"
  { TOP }
| "false"
  { BOTTOM }
| "!"
  { BANG }
| "+"
  { if is_first_line_char lexbuf then LSPLUS else PLUS }
| "-"
  { if is_first_line_char lexbuf then LSMINUS else MINUS }
| "*"
  { if is_first_line_char lexbuf then LSSTAR else STAR }
| "/"
  { DIV }
| "→"
  { ARROW }
| "=="
  { EQ }
| "!="
  { NEQ }
| "<="
  { LE }
| ">="
  { GE }
| "<"
  { LT }
| ">"
  { GT }
| "&&"
  { AND }
| "||"
  { OR }
| "="
  { ASSIGN }
| ","
  { COMMA }
| (['0'-'9']+) as i
  { INT (int_of_string i) }
| (['_''a'-'z''A'-'Z']['_''a'-'z''A'-'Z''0'-'9']*) as x
  { VAR x }
| eof
  { EOF }
| _
  { let cpos = lexbuf.lex_curr_p in
    Printf.eprintf "Line %d, character %d: Invalid character.\n"
        cpos.pos_lnum (cpos.pos_cnum - cpos.pos_bol) ;
    Pervasives.exit 1 }

