/* Copyright (c) 2015, Thibaut Girka
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY AUTHORS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE. */

%{

open CorrelatingProgram.CPStr.LangDefs
open Common

%}

%token LSSTAR LSPLUS LSMINUS

%token ASSIGN
%token ARROW
%token PLUS MINUS STAR DIV
%token EQ NEQ LE LT GE GT BANG TOP BOTTOM AND OR

%token LPAREN RPAREN LBRACKET RBRACKET LBRACE RBRACE LABRACKET RABRACKET
%token WHILE IF ELSE EOF SEMICOLON
%token BREAK CONTINUE

%token CONSTMOD VOIDTYPE INTTYPE
%token COMMA

%token <int> INT
%token <string> VAR

%nonassoc ASSIGN
%left OR
%left AND
%nonassoc EQ NEQ LE LT GE GT
%left PLUS MINUS
%left STAR DIV
%right BANG
%right SEMICOLON

%type <CorrelatingProgram.CPStr.LangDefs.cmd_diff> diffmap_prog
%type <string * (Common.decl list) * (Common.decl list) * CorrelatingProgram.CPStr.LangDefs.cmd> function_cmd
%type <string * (Common.decl list) * ((Common.decl list) * (Common.decl list)) * CorrelatingProgram.CPStr.LangDefs.cmd_diff> function_diffmap
%start function_cmd
%start function_diffmap

%%

program:
|
  { Leaf Skip }
| command SEMICOLON program
  { Seq ($1, $3) }

lval:
| VAR
  { Var $1 }
| lval LBRACKET exp RBRACKET
  { ArrayAccess ($1, $3) }

leaf:
| lval ASSIGN exp
  { Assign ($1, $3) }
| BREAK
  { Break }
| CONTINUE
  { Continue }

command:
| leaf
  { Leaf $1 }
| IF LPAREN bexp RPAREN command opt_semicolon ELSE command
  { IfThenElse ($3, $5, $8) }
| WHILE LPAREN bexp RPAREN command
  { While ($3, $5) }
| LBRACE program RBRACE
  { $2 }

exp:
| LPAREN exp RPAREN
  { $2 }
| INT
  { Int $1 }
| exp PLUS exp
  { Sum ($1, $3) }
| exp MINUS exp
  { Diff ($1, $3) }
| exp STAR exp
  { Mult ($1, $3) }
| exp DIV exp
  { Div ($1, $3) }
| lval
  { Lval $1 }

bexp:
| LPAREN bexp RPAREN
  { $2 }
| TOP
  { BTrue }
| BOTTOM
  { BFalse }
| BANG bexp
  { BNot $2 }
| bexp AND bexp
  { BAnd ($1, $3) }
| bexp OR bexp
  { BOr ($1, $3) }
| exp LE exp
  { BLE ($1, $3) }
| exp EQ exp
  { BEQ ($1, $3) }
| exp NEQ exp
  { BNEQ ($1, $3) }
| exp GE exp
  { BGE ($1, $3) }
| exp LT exp
  { BLT ($1, $3) }
| exp GT exp
  { BGT ($1, $3) }


/* Diff-mappings */
diffmap_prog:
|
  { LeafSubst (Skip, Skip) }
| EOF
  { LeafSubst (Skip, Skip) }
| LSSTAR diffmap_prog
  { CPoint $2 }
| diffmap SEMICOLON diffmap_prog
  { SeqMut ($1, $3) }
| LSPLUS command SEMICOLON diffmap_prog
  { SeqInsL ($2, $4) }
| LSMINUS command SEMICOLON diffmap_prog
  { SeqDelL ($2, $4) }

lval_diff:
| lval ARROW lval
  { ($1, $3) }
| LABRACKET lval RABRACKET ARROW LABRACKET lval RABRACKET
  { ($2, $6) }

exp_diff:
| exp ARROW exp
  { ($1, $3) }
| LABRACKET exp RABRACKET ARROW LABRACKET exp RABRACKET
  { ($2, $6) }

bexp_diff:
| bexp ARROW bexp
  { ($1, $3) }
| LABRACKET bexp RABRACKET ARROW LABRACKET bexp RABRACKET
  { ($2, $6) }

opt_semicolon:
| {}
| SEMICOLON {}

diffmap:
| leaf
  { LeafSubst ($1, $1) }
| leaf ARROW leaf
  { LeafSubst ($1, $3) }
| LBRACE diffmap_prog RBRACE
  { $2 }
| IF LPAREN bexp_diff RPAREN diffmap opt_semicolon ELSE diffmap
  { IfMut ($3, $5, $8) }
| WHILE LPAREN bexp_diff RPAREN diffmap
  { WhileMut ($3, $5) }
| LSPLUS LABRACKET IF LPAREN bexp RPAREN command opt_semicolon ELSE RABRACKET diffmap
  { ITEAddIf ($5, $7, $11) }
| LSMINUS LABRACKET IF LPAREN bexp RPAREN command opt_semicolon ELSE RABRACKET diffmap
  { ITEDelIf ($5, $7, $11) }
| LSPLUS LABRACKET IF LPAREN bexp RPAREN RABRACKET diffmap opt_semicolon LSPLUS LABRACKET ELSE command RABRACKET
  { ITEAddElse ($5, $8, $13) }
| LSMINUS LABRACKET IF LPAREN bexp RPAREN RABRACKET diffmap opt_semicolon LSMINUS LABRACKET ELSE command RABRACKET
  { ITEDelElse ($5, $8, $13) }
| LSPLUS LABRACKET WHILE LPAREN bexp RPAREN RABRACKET diffmap
  { WhileAdd ($5, $8) }
| LSMINUS LABRACKET WHILE LPAREN bexp RPAREN RABRACKET diffmap
  { WhileDel ($5, $8) }
/* Simplified syntax for assignments */
| LPAREN lval_diff RPAREN ASSIGN LPAREN exp_diff RPAREN
  { let (x, y) = $2
    and (e, e') = $6 in
    LeafSubst (Assign (x, e), Assign (y, e')) }
| LPAREN lval_diff RPAREN ASSIGN exp
  { let (x, y) = $2 in
    LeafSubst (Assign (x, $5), Assign (y, $5)) }
| lval ASSIGN LPAREN exp_diff RPAREN
  { let (e, e') = $4 in
    LeafSubst (Assign ($1, e), Assign ($1, e')) }
/* Simplified syntax for identity */
| IF LPAREN bexp RPAREN diffmap opt_semicolon ELSE diffmap
  { IfMut (($3, $3), $5, $8) }
| WHILE LPAREN bexp RPAREN diffmap
  { WhileMut (($3, $3), $5) }

param:
| INTTYPE VAR
  { ($2, IntType) }
| CONSTMOD INTTYPE STAR VAR
  { ($4, ArrayType (IntType, 0)) }

params:
|
  { [] }
| param
  { [$1] }
| param COMMA params
  { $1::$3 }

decl:
| INTTYPE VAR
  { ($2, IntType) }
| decl LBRACKET INT RBRACKET
  { let (n, t) = $1 in
    (n, ArrayType (t, $3)) }

decls:
|
  { [] }
| decl SEMICOLON decls
  { $1::$3 }

prog_diff:
| diffmap_prog
  { ([], [], $1) }
| decl SEMICOLON prog_diff
  { let (a, b, d) = $3 in
    ($1::a, $1::b, d) }
| LSPLUS decl SEMICOLON prog_diff
  { let (a, b, d) = $4 in
    (a, $2::b, d) }
| LSMINUS decl SEMICOLON prog_diff
  { let (a, b, d) = $4 in
    ($2::a, b, d) }

function_diffmap:
| VOIDTYPE VAR LPAREN params RPAREN LBRACE prog_diff RBRACE
  { let (a, b, d) = $7 in
    ($2, $4, (a, b), d) }

function_cmd:
| VOIDTYPE VAR LPAREN params RPAREN LBRACE decls program RBRACE
  { ($2, $4, $7, $8) }

