(* Copyright (c) 2015, Thibaut Girka
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY AUTHORS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE. *)

module T = Map.Make(struct type t = Z.t * Z.t let compare = compare end);;

let s0 x = Z.shift_left x 1
let s1 x = Z.succ (Z.shift_left x 1)
let epsilon = Z.one


open CorrelatingProgram;;
open CorrelatingProgram.CPStr.LangDefs;;

open Common;;

(* Those constants are arbitrary and could be optimized by analyzing the
   results on a given corpus.
   The idea is that inserting or deleting a If/Then/Else costs slightly more
   than a leaf insertion/deletion because it creates new execution paths,
   while inserting or deleting a while loop costs much more as precision
   might be lost in widening. *)
let leaf_insertion_deletion_cost = 1;;
let ite_insertion_deletion_cost = 5;;
let while_insertion_deletion_cost = 50;;

let rec insertion_cost = function
  | Leaf Skip -> 0
  | Leaf Break | Leaf Continue | Leaf (Assign (_, _)) ->
    leaf_insertion_deletion_cost
  | Seq (c, c') ->
    (insertion_cost c) + (insertion_cost c')
  | IfThenElse (b, c, c') ->
    (insertion_cost c) + (insertion_cost c') + ite_insertion_deletion_cost
  | While (b, c) ->
    (insertion_cost c) + while_insertion_deletion_cost
let deletion_cost = insertion_cost


let rec diffmap_cost = function
  | LeafSubst (a, a') when a = a' -> 0
  | LeafSubst (a, a') ->
    (deletion_cost (Leaf a)) + (insertion_cost (Leaf a'))
  | SeqMut (cd, cd') -> (diffmap_cost cd) + (diffmap_cost cd')
  | SeqDelL (c, cd) | SeqDelR (cd, c) -> (deletion_cost c) + (diffmap_cost cd)
  | SeqInsL (c, cd) | SeqInsR (cd, c) -> (insertion_cost c) + (diffmap_cost cd)
  | IfMut ((b, b'), cd, cd') ->
    (if b = b' then 0 else 1) + (diffmap_cost cd) + (diffmap_cost cd')
  | WhileMut ((b, b'), cd) ->
    (if b = b' then 0 else 1) + (diffmap_cost cd)
  | ITEAddIf (_, c, cd) | ITEAddElse (_, cd, c) ->
    (diffmap_cost cd) + (insertion_cost c) + ite_insertion_deletion_cost
  | ITEDelIf (_, c, cd) | ITEDelElse (_, cd, c) ->
    (diffmap_cost cd) + (deletion_cost c) + ite_insertion_deletion_cost
  | WhileAdd (_, cd) | WhileDel (_, cd) ->
    (diffmap_cost cd) + while_insertion_deletion_cost
  | CPoint cd -> diffmap_cost cd


let minsolution = function
  | x::xs ->
    let _, y = List.fold_left (fun (cost, cd) cd' ->
      let cost' = diffmap_cost cd' in
      if cost <= cost' then (cost, cd) else (cost', cd'))
      (diffmap_cost x, x) xs
    in y
  | _ -> failwith "Empty list!"


let rec id_diff = function
  | Leaf a -> LeafSubst (a, a)
  | IfThenElse (b, c1, c2) -> IfMut ((b, b), id_diff c1, id_diff c2)
  | While (b, c) -> WhileMut ((b, b), id_diff c)
  | Seq (c1, c2) -> SeqMut (id_diff c1, id_diff c2)


let diff c c' =
  let memo = ref T.empty in
  let rec aux' (c, c') (p1, p2) =
    let aux (c, c') key =
      try
        T.find key (!memo)
      with Not_found ->
        let cd = aux' (c, c') key in
        memo := T.add key cd (!memo) ;
        cd
  in
  if c = c' then id_diff c else
  match (c, c') with
  (* Leaves *)
  | (Leaf a, Leaf a') -> LeafSubst (a, a')
  (* I'm not 100% sure about those two... but they'd speed up things... *)
  | (Seq (c1, c2), Seq (c1', c2')) when c1 = c1' ->
    SeqMut (id_diff c1, aux (c2, c2') (s1 p1, s1 p2))
  | (Seq (c1, c2), Seq (c1', c2')) when c2 = c2' ->
    SeqMut (aux (c1, c1') (s0 p1, s0 p2), id_diff c2)
  (* Here comes the fun stuff... *)
  | (Seq (c1, c2), Seq (c1', c2')) ->
    minsolution [
      (let cd1 = aux (c1, c1') (s0 p1, s0 p2)
       and cd2 = aux (c2, c2') (s1 p1, s1 p2) in
       SeqMut (cd1, cd2)) ;
      (let cd = aux (c2, c') (s1 p1, p2) in
       SeqDelL (c1, cd)) ;
      (let cd = aux (c1, c') (s0 p1, p2) in
       SeqDelR (cd, c2)) ;
      (let cd = aux (c, c2') (p1, s1 p2) in
       SeqInsL (c1', cd)) ;
      (let cd = aux (c, c1') (p1, s0 p2) in
       SeqInsR (cd, c2'))
    ]
  | (Seq (c1, c2), IfThenElse (b, c1', c2')) ->
    minsolution [
      (let cd = aux (c2, c') (s1 p1, p2) in
       SeqDelL (c1, cd)) ;
      (let cd = aux (c1, c') (s0 p1, p2) in
       SeqDelR (cd, c2)) ;
      (let cd = aux (c, c2') (p1, s1 p2) in
       ITEAddIf (b, c1', cd)) ;
      (let cd = aux (c, c1') (p1, s0 p2) in
       ITEAddElse (b, cd, c2'))
    ]
  | (Seq (c1, c2), While (b, c'')) ->
    minsolution [
      (let cd = aux (c2, c') (s1 p1, p2) in
       SeqDelL (c1, cd)) ;
      (let cd = aux (c1, c') (s0 p1, p2) in
       SeqDelR (cd, c2)) ;
      (let cd = aux (c, c'') (p1, s1 p2) in
       WhileAdd (b, cd))
    ]
  | (IfThenElse (b, c1, c2), Seq (c1', c2')) ->
    minsolution [
      (let cd = aux (c2, c') (s1 p1, p2) in
       ITEDelIf (b, c1, cd)) ;
      (let cd = aux (c1, c') (s0 p1, p2) in
       ITEDelElse (b, cd, c2)) ;
      (let cd = aux (c, c2') (p1, s1 p2) in
       SeqInsL (c1', cd)) ;
      (let cd = aux (c, c1') (p1, s0 p2) in
       SeqInsR (cd, c2'))
    ]
  | (IfThenElse (b, c1, c2), IfThenElse (b', c1', c2')) ->
    minsolution [
      (let cd1 = aux (c1, c1') (s0 p1, s0 p2)
       and cd2 = aux (c2, c2') (s1 p1, s1 p2) in
       IfMut ((b, b'), cd1, cd2)) ;
      (let cd = aux (c2, c') (s1 p1, p2) in
       ITEDelIf (b, c1, cd)) ;
      (let cd = aux (c1, c') (s0 p1, p2) in
       ITEDelElse (b, cd, c2)) ;
      (let cd = aux (c, c2') (p1, s1 p2) in
       ITEAddIf (b', c1', cd)) ;
      (let cd = aux (c, c1') (p1, s0 p2) in
       ITEAddElse (b', cd, c2'))
    ]
  | (IfThenElse (b, c1, c2), While (b', c'')) ->
    minsolution [
      (let cd = aux (c2, c') (s1 p1, p2) in
       ITEDelIf (b, c1, cd)) ;
      (let cd = aux (c1, c') (s0 p1, p2) in
       ITEDelElse (b, cd, c2)) ;
      (let cd = aux (c, c'') (p1, s1 p2) in
       WhileAdd (b', cd))
    ]
  | (While (b, c1), Seq (c1', c2')) ->
    minsolution [
      (let cd = aux (c1, c') (s1 p1, p2) in
       WhileDel (b, cd)) ;
      (let cd = aux (c, c2') (p1, s1 p2) in
       SeqInsL (c1', cd)) ;
      (let cd = aux (c, c1') (p1, s0 p2) in
       SeqInsR (cd, c2'))
    ]
  | (While (b, c1), IfThenElse (b', c1', c2')) ->
    minsolution [
      (let cd = aux (c1, c') (s1 p1, p2) in
       WhileDel (b, cd)) ;
      (let cd = aux (c, c2') (p1, s1 p2) in
       ITEAddIf (b', c1', cd)) ;
      (let cd = aux (c, c1') (p1, s0 p2) in
       ITEAddElse (b', cd, c2'))
    ]
  | (While (b, c1), While (b', c1')) ->
    minsolution [
      (let cd1 = aux (c1, c1') (s1 p1, s1 p2) in
       WhileMut ((b, b'), cd1)) ;
      (let cd = aux (c1, c') (s1 p1, p2) in
       WhileDel (b, cd)) ;
      (let cd = aux (c, c1') (p1, s1 p2) in
       WhileAdd (b', cd))
    ]
  | (While (b, c1), _) ->
    let cd = aux (c1, c') (s1 p1, p2) in
    WhileDel (b, cd)
  | (_, While (b', c1')) ->
    let cd = aux (c, c1') (p1, s1 p2) in
    WhileAdd (b', cd)
  | (IfThenElse (b, c1, c2), _) ->
    minsolution [
      (let cd = aux (c2, c') (s1 p1, p2) in
       ITEDelIf (b, c1, cd)) ;
      (let cd = aux (c1, c') (s0 p1, p2) in
       ITEDelElse (b, cd, c2))
    ]
  | (_, IfThenElse (b', c1', c2')) ->
    minsolution [
      (let cd = aux (c, c2') (p1, s1 p2) in
       ITEAddIf (b', c1', cd)) ;
      (let cd = aux (c, c1') (p1, s0 p2) in
       ITEAddElse (b', cd, c2'))
    ]
  | (Seq (c1, c2), _) ->
    minsolution [
      (let cd = aux (c2, c') (s1 p1, p2) in
       SeqDelL (c1, cd)) ;
      (let cd = aux (c1, c') (s0 p1, p2) in
       SeqDelR (cd, c2))
    ]
  | (_, Seq (c1', c2')) ->
    minsolution [
      (let cd = aux (c, c2') (p1, s1 p2) in
       SeqInsL (c1', cd)) ;
      (let cd = aux (c, c1') (p1, s0 p2) in
       SeqInsR (cd, c2'))
    ]
  in
  let cd = aux' (c, c') (epsilon, epsilon) in
  assert ((__U03a0___U2080_ cd) = c) ;
  assert ((__U03a0___U2081_ cd) = c') ;
  cd

module DeclSet = Set.Make (struct type t = decl let compare = compare end)

let diff_decls decls decls' =
  let decls = List.fold_left (fun a x -> DeclSet.add x a) DeclSet.empty decls
  and decls' = List.fold_left (fun a x -> DeclSet.add x a) DeclSet.empty decls' in
  let common = DeclSet.inter decls decls' in
  let only_decls = DeclSet.diff decls common
  and only_decls' = DeclSet.diff decls' common in
  (common, only_decls, only_decls')

let rec cmd_to_c' (x : string) (indent : string) (c : cmd) : string =
  match c with
  | Seq (c, c') ->
    (cmd_to_c' x indent c) ^ (cmd_to_c' x indent c')
  | Leaf Skip -> ""
  | IfThenElse (b, c, c') -> x ^ indent ^ "if (" ^ (bexp_to_c b) ^ ") {\n" ^
    (cmd_to_c ("   " ^ indent) c) ^ " " ^ indent ^ "} else {\n" ^
    (cmd_to_c ("   " ^ indent) c') ^ " " ^ indent ^ "};\n"
  | While (b, c) -> x ^ indent ^ "while (" ^ (bexp_to_c b) ^ ") {\n" ^
    (cmd_to_c ("   " ^ indent) c) ^ " " ^ indent ^ "};\n"
  | _ -> x ^ (cmd_to_c indent c)

let print_diff indent cd =
  let rec print_diff indent = function
  | LeafSubst (Skip, Skip) -> ""
  | LeafSubst (Skip, _) | LeafSubst (_, Skip) -> failwith "Incorrect diff!"
  | LeafSubst (a, b) when a = b -> " " ^ indent ^ (leaf_to_c a) ^ ";\n"
  | LeafSubst (Assign (x, e), Assign (y, e')) when x = y ->
    " " ^ indent ^ (lval_to_c x) ^ " = " ^ "(⟨" ^ (exp_to_c e) ^ "⟩ → ⟨" ^ (exp_to_c e') ^ "⟩);\n"
  | LeafSubst (Assign (x, e), Assign (y, e')) when e = e' ->
    " " ^ indent ^ "(⟨" ^ (lval_to_c x) ^ "⟩ → ⟨" ^ (lval_to_c y) ^ "⟩) = " ^ (exp_to_c e) ^ ";\n"
  | LeafSubst (a, b) -> " " ^ indent ^ (leaf_to_c a) ^ " → " ^ (leaf_to_c b) ^ ";\n"
  | SeqDelL (c, cd) ->
    (cmd_to_c' "-" indent c) ^
    (print_diff indent cd)
  | SeqInsL (c, cd) ->
    (cmd_to_c' "+" indent c) ^
    (print_diff indent cd)
  | SeqMut (cd, cd') ->
    (print_diff indent cd) ^
    (print_diff indent cd')
  | IfMut ((b, b'), cd, cd') ->
    let bexp_part = if b = b' then bexp_to_c b else "⟨" ^ (bexp_to_c b) ^ "⟩ → ⟨" ^ (bexp_to_c b') ^ "⟩" in
    indent ^ " if (" ^ bexp_part ^ ") {\n" ^
    (print_diff (indent ^ "  ") cd) ^ " " ^ indent ^ "} else {\n" ^
    (print_diff (indent ^ "  ") cd') ^ " " ^ indent ^ "};\n"
  | WhileMut ((b, b'), cd) ->
    let bexp_part = if b = b' then bexp_to_c b else "⟨" ^ (bexp_to_c b) ^ "⟩ → ⟨" ^ (bexp_to_c b') ^ "⟩" in
    indent ^ " while (" ^ bexp_part ^ ") {\n" ^
    (print_diff (indent ^ "  ") cd) ^ " " ^ indent ^ "};\n"
  | WhileAdd (b, cd) ->
    "+" ^ indent ^ "⟨while (" ^ (bexp_to_c b) ^ ")⟩ {\n" ^
    (print_diff (indent ^ "  ") cd) ^ " " ^ indent ^ "};\n"
  | WhileDel (b, cd) ->
    "-" ^ indent ^ "⟨while (" ^ (bexp_to_c b) ^ ")⟩ {\n" ^
    (print_diff (indent ^ "  ") cd) ^ " " ^ indent ^ "};\n"
  | ITEAddIf (b, c, cd') ->
    let bexp_part = bexp_to_c b in
    "+" ^ indent ^ "⟨if (" ^ bexp_part ^ ") {\n" ^
    (cmd_to_c (indent ^ "   ") c) ^ " " ^ indent ^ "} else⟩ {\n" ^
    (print_diff (indent ^ "  ") cd') ^ " " ^ indent ^ "};\n"
  | ITEDelIf (b, c, cd') ->
    let bexp_part = bexp_to_c b in
    "-" ^ indent ^ "⟨if (" ^ bexp_part ^ ") {\n" ^
    (cmd_to_c (indent ^ "   ") c) ^ " " ^ indent ^ "} else⟩ {\n" ^
    (print_diff (indent ^ "  ") cd') ^ " " ^ indent ^ "};\n"
  | ITEAddElse (b, cd, c) ->
    let bexp_part = bexp_to_c b in
    "+" ^ indent ^ "⟨if (" ^ bexp_part ^ ")⟩ {\n" ^
    (print_diff (indent ^ "  ") cd) ^ " " ^ indent ^ "} ⟨else {\n" ^
    (cmd_to_c (indent ^ "   ") c) ^ " " ^ indent ^ "⟩};\n"
  | ITEDelElse (b, cd, c) ->
    let bexp_part = bexp_to_c b in
    "-" ^ indent ^ "⟨if (" ^ bexp_part ^ ")⟩ {\n" ^
    (print_diff (indent ^ "  ") cd) ^ " " ^ indent ^ "} ⟨else {\n" ^
    (cmd_to_c (indent ^ "   ") c) ^ " " ^ indent ^ "⟩};\n"
  | CPoint cd ->
    "*\n" ^ (print_diff indent cd)
  | _ -> failwith "Incorrect diff!"
  in print_diff indent (CPStr.simplify_diff cd)


let print_decls_diff (common, only_decls, only_decls') =
  let lines = [] in
  let lines = DeclSet.fold (fun x a -> ("   " ^ (decl_to_c x) ^ ";")::a) common lines in
  let lines = DeclSet.fold (fun x a -> ("-  " ^ (decl_to_c x) ^ ";")::a) only_decls lines in
  let lines = DeclSet.fold (fun x a -> ("+  " ^ (decl_to_c x) ^ ";")::a) only_decls' lines in
  (String.concat "\n" lines)


let diff_to_string name params decls_diff d =
" void " ^ name ^ "(" ^ (String.concat ", " (List.map param_to_c params)) ^ ") {
" ^ (print_decls_diff decls_diff) ^ "
" ^ (print_diff "  " d) ^ "
 }
"

let compute_diff orig_path mut_path diff_path =
  let in_channel = open_in orig_path in
  let lex = Lexing.from_channel in_channel in
  let in_channel' = open_in mut_path in
  let lex' = Lexing.from_channel in_channel' in
  try
    let (name, params, decls, c) = Parser.function_cmd Lexer.main lex in
    let () = close_in in_channel in
    let (name', params', decls', c') = Parser.function_cmd Lexer.main lex' in
    let () = close_in in_channel' in
    assert ((name, params) = (name', params')) ;
    let decls_diff = diff_decls decls decls' in
    let cd = diff c c' in
    let out_channel = open_out diff_path in
    let () = output_string out_channel (diff_to_string name params decls_diff cd) in
    close_out out_channel

  with Parsing.Parse_error ->
    let cpos = lex.Lexing.lex_curr_p in
    Printf.eprintf "Line %d, character %d: Syntax error.\n"
        cpos.Lexing.pos_lnum (cpos.Lexing.pos_cnum - cpos.Lexing.pos_bol);
    Pervasives.exit 1


let main (argv : string array) =
  compute_diff (argv.(1)) (argv.(2)) (argv.(3))


let () = main (Sys.argv);;
